#include "trace.h"
#include "Processor/instance.h"
#include "Processor/state.h"
#include "MemSys/simsys.h"
#include <stdarg.h>
#include <inttypes.h>

#ifdef TRACE
static FILE *tracefile = NULL;
static int trace_start_cycle = 0;

static int
vtracef_plain (const char *format, va_list args)
{
	if (tracefile && YS__Simtime > trace_start_cycle)
	{
		return vfprintf (tracefile, format, args);
	}
	else
	{
		return 0;
	}
}

void 
trace_set_file (FILE *f)
{
	tracefile = f;
}

void
trace_set_start_cycle (int cycle)
{
	trace_start_cycle = cycle;
}

int
tracef (const char *system, int system_id, const char *format, ...)
{
	if (tracefile && YS__Simtime > trace_start_cycle)
	{
		fprintf (tracefile, system, system_id);
		fputs (": ", tracefile);
		va_list args;
		va_start (args, format);
		int ret = vtracef_plain (format, args);
		va_end (args);
		return ret;
	}
	else
	{
		return 0;
	}
}

int 
tracef_plain (const char *format, ...)
{
	if (tracefile && YS__Simtime > trace_start_cycle)
	{
		va_list args;
		va_start (args, format);
		int ret = vfprintf (tracefile, format, args);
		va_end (args);
		return ret;
	}
	else
	{
		return 0;
	}
}

int
tracef_inst (state *proc, instance *inst)
{
	if (YS__Simtime > trace_start_cycle)
	{
		instr *instrn = inst->code;	
		int ret = tracef_plain ("%10d %6d %7x: %-12.11s%4d%4d%4d%4d%4d%4d%4d%4d%4d%8d ", 
			proc->curr_cycle, 
			inst->tag, 
			inst->pc, 
			inames[instrn->instruction], 
			instrn->rd, 
			inst->prd,
			instrn->rcc,
			instrn->rs1,
			inst->prs1,
			instrn->rs2,
			inst->prs2,
			instrn->aux1,
			instrn->aux2, 
			instrn->imm);
		switch (instrn->rs1_regtype)
		{
		case REG_INT:
		case REG_INT64:
			tracef_plain ("%8x ", inst->rs1vali);
			break;
		case REG_FP:
			break;
		case REG_FPHALF:
			break;
		case REG_INTPAIR:
			break;
		default:
			break;
		}
	
		switch (instrn->rs2_regtype)
		{
		case REG_INT:
		case REG_INT64:
			tracef_plain ("%8x ", inst->rs2vali);
			break;
		case REG_FP:
			break;
		case REG_FPHALF:
			break;
		case REG_INTPAIR:
			break;
		default:
			break;
		}
	
		switch (instrn->rd_regtype)
		{
		case REG_INT:
		case REG_INT64:
			tracef_plain ("%8x ", inst->rdvali);
			break;
		case REG_FP:
			break;
		case REG_FPHALF:
			break;
		case REG_INTPAIR:
			break;
		default:
			break;
		}

		tracef_plain ("\n");
		return ret;
	}
	else
	{
		return 0;
	}
}

void
tracef_registers (state *proc, const char *tip)
{
	if (YS__Simtime > trace_start_cycle)
	{
		for (int i = 0; i < NO_OF_LOGICAL_INT_REGISTERS+MAX_MAX_ACTIVE_NUMBER; ++i)
		{
			if (proc->physical_int_reg_file[i] != proc->debug_int_regs[i])
			{
				tracef ("proc%d", proc->proc_id, "%4d: %08x %s\n", i, proc->physical_int_reg_file[i], tip);
				proc->debug_int_regs[i] = proc->physical_int_reg_file[i];
			}
		}
		for (int i = 0; i < NO_OF_LOGICAL_FP_REGISTERS+MAX_MAX_ACTIVE_NUMBER; ++i)
		{
			uint64_t *p = (uint64_t *) proc->physical_fp_reg_file;
			if (proc->physical_fp_reg_file[i] != proc->debug_fp_regs[i])
			{
				proc->debug_fp_regs[i] = proc->physical_fp_reg_file[i];
				tracef ("proc%d", proc->proc_id, "%4d: %016llx %f %s\n", i, p[i], proc->debug_fp_regs[i], tip);
			}
		}
	}
}
#endif
